/*******************************************************************
 * Copyright (C) 2000 SSI. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/
/**************************************************
 * RCSId: $Id: calc_sunmoon.c,v 1.2 2016/03/25 15:41:56 ralblas Exp $
 *
 * Calculate current observation point for polar satellites 
 * Project: SSI
 * Author: R. Alblas
 *
 * History: 
 * $Log: calc_sunmoon.c,v $
 * Revision 1.2  2016/03/25 15:41:56  ralblas
 * _
 *
 * Revision 1.1  2015/11/18 18:33:49  ralblas
 * Initial revision
 *
 **************************************************/
#include <math.h>
#include <float.h>
#include <string.h>
#include "sattrack.h"
#include "xtrack.h"
#include "xtrack_func.h"

/********************************************************
 * Calc. sun
 ********************************************************/
  //# Astronomer's almanach time is the number of 
  //# days since (noon, 1 January 2000)
static float astronomersAlmanacTime(struct tm tm)
{
  struct tm tm1;
  memset(&tm1,0,sizeof(tm1));
  tm1.tm_mday=1;
  tm1.tm_mon=0;
  tm1.tm_year=2000-1900;
  tm1.tm_hour=12;
  long tx=mktime_ntz(&tm);
  long ty=mktime_ntz(&tm1);
  float dt=difftime(tx,ty);
  return dt/(24.*3600.);
}

static float hourOfDay(struct tm tm)
{
  float h=tm.tm_hour+tm.tm_min/60.+tm.tm_sec/3600.;
  return h;
}

static float meanLongitudeDegrees(float time)
{
  float ret;
  ret=(280.460 + 0.9856474 * time);
  while (ret>360) ret-=360;
  while (ret<0) ret+=360;
  return ret;
}

static float meanAnomalyRadians(float time)
{
  float ret;
  ret=357.528 + 0.9856003 * time;
  while (ret>360) ret-=360;
  while (ret<0) ret+=360;
  return D2R(ret);
}

static float eclipticLongitudeRadians(float mnlong, float mnanom)
{
  float ret;
  ret=mnlong + 1.915 * sin(mnanom) + 0.020 * sin(2 * mnanom);
  while (ret>360) ret-=360;
  while (ret<0) ret+=360;
  return D2R(ret);
}

static float eclipticObliquityRadians(float time)
{
  return D2R(23.439 - 0.0000004 * time);
}

static float rightAscensionRadian(float oblqec, float eclong)
{
  float num,den,ra;
  num=cos(oblqec) * sin(eclong);
  den=cos(eclong);
  ra=atan2(num,den);
  return ra;
}

static float rightDeclinationRadians(float oblqec, float eclong)
{
  return asin(sin(oblqec) * sin(eclong));
}


static float greenwichMeanSiderealTimeHours(float time, float hour)
{
  float ret;
  ret=(6.697375 + 0.0657098242 * time + hour);
  while (ret>24) ret-=24;
  while (ret<0) ret+=24;
  return ret;
}

static float localMeanSiderealTimeRadians(float gmst, float lon)
{
  float ret=(gmst + lon / 15.);
  while (ret>24) ret-=24;
  while (ret<0) ret+=24;
  return D2R(15.*ret);
}

static float hourAngleRadians(float lmst, float ra)
{
  float ret;
  ret=lmst - ra + PI;
  while (ret>2.*PI) ret-=2.*PI;
  return ret-PI;
}


EPOINT calc_sun(struct tm tm)
{
  EPOINT zon;
  float time,fhour;
  float lon=0.;
  float mnlong,mnanom,eclong,oblqec;
  float ra,dec,gmst,lmst;
  float ha;
  time=astronomersAlmanacTime(tm);
  fhour = hourOfDay(tm);

  //# Ecliptic coordinates  
  mnlong = meanLongitudeDegrees(time);  
  mnanom = meanAnomalyRadians(time);
  eclong = eclipticLongitudeRadians(mnlong, mnanom);   
  oblqec = eclipticObliquityRadians(time);

  //# Celestial coordinates
  ra =rightAscensionRadian(oblqec, eclong);
  dec = rightDeclinationRadians(oblqec, eclong);

  //# Local coordinates
  gmst = greenwichMeanSiderealTimeHours(time, fhour) ; 
  lmst = localMeanSiderealTimeRadians(gmst, lon);

  //# Hour angle
  ha = hourAngleRadians(lmst, ra);

  zon.lon=-1.*ha;
  zon.lat=dec;
  return zon;
}



 int T45AD[] = {0, 2, 2, 0, 0, 0, 2, 2, 2, 2,
                         0, 1, 0, 2, 0, 0, 4, 0, 4, 2,
                         2, 1, 1, 2, 2, 4, 2, 0, 2, 2,
                         1, 2, 0, 0, 2, 2, 2, 4, 0, 3,
                         2, 4, 0, 2, 2, 2, 4, 0, 4, 1,
                         2, 0, 1, 3, 4, 2, 0, 1, 2, 2};

  int  T45AM[] = {0,  0,  0,  0,  1,  0,  0, -1,  0, -1,
                         1,  0,  1,  0,  0,  0,  0,  0,  0,  1,
                         1,  0,  1, -1,  0,  0,  0,  1,  0, -1,
                         0, -2,  1,  2, -2,  0,  0, -1,  0,  0,
                         1, -1,  2,  2,  1, -1,  0,  0, -1,  0,
                         1,  0,  1,  0,  0, -1,  2,  1,  0,  0};

   int T45AMP[] = { 1, -1,  0,  2,  0,  0, -2, -1,  1,  0,
                          -1,  0,  1,  0,  1,  1, -1,  3, -2, -1,
                           0, -1,  0,  1,  2,  0, -3, -2, -1, -2,
                           1,  0,  2,  0, -1,  1,  0, -1,  2, -1,
                           1, -2, -1, -1, -2,  0,  1,  4,  0, -2,
                           0,  2,  1, -2, -3,  2,  1, -1,  3, -1};

   int T45AF[]  = { 0,  0,  0,  0,  0,  2,  0,  0,  0,  0,
                           0,  0,  0, -2,  2, -2,  0,  0,  0,  0,
                           0,  0,  0,  0,  0,  0,  0,  0,  2,  0,
                           0,  0,  0,  0,  0, -2,  2,  0,  2,  0,
                           0,  0,  0,  0,  0, -2,  0,  0,  0,  0,
                          -2, -2,  0,  0,  0,  0,  0,  0,  0, -2};

   int T45AL[] = {6288774, 1274027, 658314, 213618, -185116,
                         -114332,   58793,  57066,  53322,   45758,
                          -40923,  -34720, -30383,  15327,  -12528,
                           10980,   10675,  10034,   8548,   -7888,
                           -6766,   -5163,   4987,   4036,    3994,
                            3861,    3665,  -2689,  -2602,    2390,
                           -2348,    2236,  -2120,  -2069,    2048,
                           -1773,   -1595,   1215,  -1110,    -892,
                            -810,     759,   -713,   -700,     691,
                             596,     549,    537,    520,    -487,
                            -399,    -381,    351,   -340,     330,
                             327,    -323,    299,    294,       0};

   int T45AR[] = {-20905355, -3699111, -2955968, -569925,   48888,
                             -3149,   246158,  -152138, -170733, -204586,
                           -129620,   108743,   104755,   10321,       0,
                             79661,   -34782,   -23210,  -21636,   24208,
                             30824,    -8379,   -16675,  -12831,  -10445,
                            -11650,    14403,    -7003,       0,   10056,
                              6322,    -9884,     5751,       0,   -4950,
                              4130,        0,    -3958,       0,    3258,
                              2616,    -1897,    -2117,    2354,       0,
                                 0,    -1423,    -1117,   -1571,   -1739,
                                 0,    -4421,        0,       0,       0,
                                 0,     1165,        0,       0,    8752};

// Meeus table 45B latitude of the moon

   int T45BD[] = {0, 0, 0, 2, 2, 2, 2, 0, 2, 0,
                         2, 2, 2, 2, 2, 2, 2, 0, 4, 0,
                         0, 0, 1, 0, 0, 0, 1, 0, 4, 4,
                         0, 4, 2, 2, 2, 2, 0, 2, 2, 2,
                         2, 4, 2, 2, 0, 2, 1, 1, 0, 2,
                         1, 2, 0, 4, 4, 1, 4, 1, 4, 2};

   int T45BM[] = { 0,  0,  0,  0,  0,  0,  0, 0,  0,  0,
                         -1,  0,  0,  1, -1, -1, -1, 1,  0,  1,
                          0,  1,  0,  1,  1,  1,  0, 0,  0,  0,
                          0,  0,  0,  0, -1,  0,  0, 0,  0,  1,
                          1,  0, -1, -2,  0,  1,  1, 1,  1,  1,
                          0, -1,  1,  0, -1,  0,  0, 0, -1, -2};

   int T45BMP[] = {0,  1, 1,  0, -1, -1,  0,  2,  1,  2,
                          0, -2, 1,  0, -1,  0, -1, -1, -1,  0,
                          0, -1, 0,  1,  1,  0,  0,  3,  0, -1,
                          1, -2, 0,  2,  1, -2,  3,  2, -3, -1,
                          0,  0, 1,  0,  1,  1,  0,  0, -2, -1,
                          1, -2, 2, -2, -1,  1,  1, -1,  0,  0};

   int T45BF[] = { 1,  1, -1, -1,  1, -1,  1,  1, -1, -1,
                         -1, -1,  1, -1,  1,  1, -1, -1, -1,  1,
                          3,  1,  1,  1, -1, -1, -1,  1, -1,  1,
                         -3,  1, -3, -1, -1,  1, -1,  1, -1,  1,
                          1,  1,  1, -1,  3, -1, -1,  1, -1, -1,
                          1, -1,  1, -1, -1, -1, -1, -1, -1,  1};

   int T45BL[] = {5128122, 280602, 277693, 173237, 55413,
                           46271,  32573,  17198,   9266,  8822,
                            8216,   4324,   4200,  -3359,  2463,
                            2211,   2065,  -1870,   1828, -1794,
                           -1749,  -1565,  -1491,  -1475, -1410,
                           -1344,  -1335,   1107,   1021,   833,
                             777,    671,    607,    596,   491,
                            -451,    439,    422,    421,  -366,
                            -351,    331,    315,    302,  -283,
                            -229,    223,    223,   -220,  -220,
                            -185,    181,   -177,    176,   166,
                            -164,    132,   -119,    115,   107};

float rev(angle){return angle-(int)(angle/360.0)*360.0;}
float sind(angle){return sin((angle*PI)/180.0);}
float cosd(angle){return cos((angle*PI)/180.0);}
float tand(angle){return tan((angle*PI)/180.0);}
float asind(c){return (180.0/PI)*asin(c);}
//float acosd(c){return (180.0/PI)*acos(c);}
float atan2d(y,x){return (180.0/PI)*atan2(y,x)-180.0*(x<0);}

#ifdef NOGTEDOEN
//float rev(a) {return a;}
static EPOINT calc_moon1(struct tm tm)
{
  EPOINT maan;
  float jd;
  jd=astronomersAlmanacTime(tm);
printf("%f\n",jd);
 float T=jd;
 float  T2=T*T;
 float  T3=T2*T;
 float  T4=T3*T;
  // Moons mean longitude L'
  float LP=218.3164477+481267.88123421*T-0.0015786*T2+T3/538841.0-T4/65194000.0;
  // Moons mean elongation
  float D=297.8501921+445267.1114034*T-0.0018819*T2+T3/545868.0-T4/113065000.0;
  // Suns mean anomaly
 float  M=357.5291092+35999.0502909*T-0.0001536*T2+T3/24490000.0;
  // Moons mean anomaly M'
 float  MP=134.9633964+477198.8675055*T+0.0087414*T2+T3/69699.0-T4/14712000.0;
  // Moons argument of latitude
 float  F=93.2720950+483202.0175233*T-0.0036539*T2-T3/3526000.0+T4/863310000.0;

  // Additional arguments
  float A1=119.75+131.849*T;
  float A2=53.09+479264.290*T;
  float A3=313.45+481266.484*T;
  float E=1-0.002516*T-0.0000074*T2;
 float  E2=E*E;
  // Sums of periodic terms from table 45.A and 45.B
  float Sl=0.0;
  float Sr=0.0;
  int i;
  for ( i=0; i<60; i++) {
    float Eterm=1;
    if (ABS(T45AM[i])==1) Eterm=E;
    if (ABS(T45AM[i])==2) Eterm=E2;
    Sl+=T45AL[i]*Eterm*sind(rev(T45AD[i]*D+T45AM[i]*M+T45AMP[i]*MP+T45AF[i]*F));
    Sr+=T45AR[i]*Eterm*cosd(rev(T45AD[i]*D+T45AM[i]*M+T45AMP[i]*MP+T45AF[i]*F));
  }
  float Sb=0.0;
  for ( i=0; i<60; i++) {
    float Eterm=1;
    if (ABS(T45BM[i])==1) Eterm=E;
    if (ABS(T45BM[i])==2) Eterm=E2;
    Sb+=T45BL[i]*Eterm*sind(rev(T45BD[i]*D+T45BM[i]*M+T45BMP[i]*MP+T45BF[i]*F));
  }
  // Additional additive terms
  Sl=Sl+3958*sind(rev(A1))+1962*sind(rev(LP-F))+318*sind(rev(A2));
  Sb=Sb-2235*sind(rev(LP))+382*sind(rev(A3))+175*sind(rev(A1-F))+
     175*sind(rev(A1+F))+127*sind(rev(LP-MP))-115*sind(rev(LP+MP));
  // geocentric longitude, latitude and distance
  float mglong=rev(LP+Sl/1000000.0);
  float mglat=rev(Sb/1000000.0);
  if (mglat > 180.0) mglat=mglat-360;
//  float mr=(385000.56+Sr/1000.0);
  // Obliquity of Ecliptic
  float obl=23.4393-3.563E-9*(jd-2451543.5);
  // RA and dec
  float ra=rev(atan2d(sind(mglong)*cosd(obl)-tand(mglat)*sin(obl),
                    cosd(mglong)))/15.0;
  float dec=rev(asind(sind(mglat)*cosd(obl)+cosd(mglat)*sind(obl)*sind(mglong)));
  if (dec > 180.0) dec=dec-360;
  maan.lon = D2R(dec);
  maan.lat = D2R(ra);
printf("%f  %f\n",R2D(dec),R2D(ra));
  return maan; //  new Array(ra,dec,mr);
}

#endif


EPOINT calc_moon(struct tm tm)
{
  EPOINT maan;
  float d;
  d=astronomersAlmanacTime(tm);
  float N = (125.1228 - 0.0529538083 * d) / 180 * PI;
  float i = 5.1454;
  float w = (318.0634 + 0.1643573223 * d) / 180 * PI;
  float e = 0.054900; //eccentricity
  float M = (115.3654 + 13.0649929509 * d) / 180.0f * PI ; //mean anomaly

  float E = M + e * sin(M) * ( 1.0 + e * cos(M) );
  E = E - ( E - e * sin(E) - M ) / ( 1 - e * cos(E) );

  float xv = ( cos(E) - e );
  float yv = ( sqrt(1.0 - e*e) * sin(E) );

  float v = atan2( yv, xv );

  float xh = ( cos(N) * cos(v+w) - sin(N) * sin(v+w) * cos(i) );
  float yh = ( sin(N) * cos(v+w) + cos(N) * sin(v+w) * cos(i) );
  float zh = ( sin(v+w) * sin(i) );

  maan.lon = atan2( yh, xh ); //longitude
  maan.lat = atan2( zh, sqrt(xh*xh+yh*yh) ); //latitude
  return maan;
}

#ifdef ZONSIMPEL
// lente: 31+28+22=81
// zomer: 81+9+30+31+22=173
// herfst:173+9+31+31+22=266
// winter: 266+8+31+30+22=357
#define LAT_KEERKR 23.439
static EPOINT calc_zon_simpel(struct tm ti)
{
  struct tm t=ti;
  EPOINT zon;
  int tjaar;
  int tdag;
  float zlon;
  float zlat;
  t.tm_year=70;
  tjaar=mktime_ntz(&t);                        // sec in jaar
  tdag=t.tm_hour*3600+t.tm_min*60+t.tm_sec;   // sec in dag
  zlon=((12.-(float)(tdag)/3600.))*90./6.;
  zlat=(float)(tjaar-(173*24*3600))/(float)((173*24*3600));
  zlat=LAT_KEERKR*cos(D2R(zlat*180.));

  zon.lon=D2R(zlon);
  zon.lat=D2R(zlat);
//printf("%f  %f\n",R2D(zon.lon),R2D(zon.lat));
  return zon;
}
#endif
